import os
import urllib.parse
from dash import Dash, State, html, dcc, callback, Output, Input
from flask import Flask
from pandas.core.generic import weakref
import plotly.express as px
import pandas as pd
from tradingo.api import Tradingo

ARCTIC_URL = os.environ.get(
    "TRADINGO_ARCTIC_URL",
    "lmdb:///home/rory/dev/tradingo-plat/data/prod/tradingo.db",
)

pd.options.plotting.backend = "plotly"


api = Tradingo(
    uri=ARCTIC_URL,
    provider="ig-trading",
    universe="im-multi-asset",
    name="etft",
)

app = Dash()

app.layout = html.Div(
    [
        html.Div(
            [
                html.H1(children="Tradingo Monitor", style={"textAlign": "center"}),
                dcc.DatePickerRange(
                    start_date=None,
                    end_date=None,
                    id="date-selection",
                ),
                dcc.Dropdown(
                    value=None,
                    options=api.instruments["im-multi-asset"]().index.to_list(),
                    id="dropdown-selection",
                    multi=True,
                ),
            ]
        ),
        html.Div(
            [
                dcc.Graph(id="z_score"),
                dcc.Graph(id="net_position"),
                dcc.Graph(id="unrealised_pnl"),
                dcc.Graph(id="total_pnl"),
                dcc.Graph(id="month-to-date"),
            ]
        ),
    ]
)


@callback(
    (
        Output("z_score", "figure"),
        Output("net_position", "figure"),
        Output("unrealised_pnl", "figure"),
        Output("total_pnl", "figure"),
        Output("month-to-date", "figure"),
    ),
    (
        Input("dropdown-selection", "value"),
        Input("date-selection", "start_date"),
        Input("date-selection", "end_date"),
    ),
)
def update_graph(
    value,
    start_date,
    end_date,
):

    def range_breaks(fig):
        fig.update_xaxes(
            rangebreaks=[
                dict(bounds=[21.5, 13.5], pattern="hour"),
                dict(bounds=["sat", "mon"]),
            ]
        )
        return fig

    value = value or None
    date = pd.Timestamp(start_date or pd.Timestamp.now()).normalize()
    end = pd.Timestamp(
        end_date or (pd.Timestamp.now().normalize() + pd.Timedelta(hours=24))
    ).normalize()
    z_score = api.signals.intraday_momentum.z_score(
        columns=value,
        date_range=(date, end),
    )

    start = z_score.index[0]
    end = z_score.index[-1] + pd.Timedelta(minutes=15)

    unrealised_pnl = (
        api.backtest.intraday.rounded.position.instrument.unrealised_pnl(
            columns=value,
            date_range=(start, end),
        )
        .ffill()
        .fillna(0.0)
        .diff()
        .cumsum()
        .fillna(0.0)
    )

    net_position = api.backtest.intraday.rounded.position.instrument.net_position(
        columns=value,
        date_range=(start, end),
    )
    total_pnl = (
        api.backtest.intraday.rounded.position.portfolio(
            date_range=(
                start,
                end,
            ),
        )
        .total_pnl.ffill()
        .fillna(0.0)
        .diff()
        .cumsum()
        .fillna(0.0)
    )

    inception = pd.Timestamp("2024-11-07 00:00:00+00:00")

    mtd = (
        api.backtest.intraday.rounded.position.portfolio(
            date_range=(
                inception,
                # end - pd.offsets.BDay(30),
                end,
            ),
        )
        .resample(pd.offsets.BDay(1))
        .last()
        .diff()
        .fillna(0)
        .cumsum()
        .total_pnl
    )

    return (
        range_breaks(z_score.plot(title="z_score")),
        range_breaks(net_position.plot(title="net_position")),
        range_breaks(unrealised_pnl.plot(title="unrealised_pnl")),
        range_breaks(total_pnl.plot(title="total_pnl")),
        mtd.plot(title="Returns since inception"),
    )


server = app.server


if __name__ == "__main__":
    app.run(
        host="0.0.0.0",
        debug=True,
    )



import pytest

from tradingo.api import Tradingo


def test_tradingo_api(tradingo: Tradingo):

    df = tradingo.prices.close()


def test_tradingo_api_with(tradingo: Tradingo):

    t = tradingo

    t.instruments.etfs.concat.prices.close.transpose(axis=1).transpose()


def test_tradingo_merge(tradingo: Tradingo):
    tradingo.portfolio.model.raw.shares.concat.prices.close(
        axis=1, keys=("position", "close")
    )


if __name__ == "__main__":
    pytest.main()



# Historical IG data may be downloaded from the following site.
# https://www.dukascopy.com/swiss/english/marketwatch/historical/
#
import argparse
import logging
from typing import List, Optional
import dateutil.tz
from collections import defaultdict
import re
import os
from pathlib import Path

from arcticdb import Arctic, LibraryOptions

import pandas as pd

from tradingo.signals import symbol_provider, symbol_publisher
from tradingo import sampling


logger = logging.getLogger(__name__)


ASSET_MAPPING = {
    "USA500.IDXUSD": "IX.D.SPTRD.IFS.IP",
    "GAS.CMDUSD": "CC.D.NG.UMP.IP",
    "BRENT.CMDUSD": "CC.D.LCO.UMP.IP",
    "USTBOND.TRUSD": "IR.D.10YEAR100.FWM2.IP",
    "COCOA.CMDUSD": "CC.D.CC.UMP.IP",
}

MULTIPLIERS = {
    "USA500.IDXUSD": 1,
    "GAS.CMDUSD": 1000,
    "BRENT.CMDUSD": 100,
    "USTBOND.TRUSD": 100,
    "COCOA.CMDUSD": 1,
}


def cli_app():

    app = argparse.ArgumentParser("load-historical-data")

    app.add_argument("--path", type=Path, required=True, nargs="+")
    app.add_argument("--arctic-uri", required=True)
    app.add_argument("--dry-run", action="store_true")
    app.add_argument("--universe", required=True)
    app.add_argument("--provider", required=True)
    app.add_argument("--clean", action="store_true")
    app.add_argument("--end-date", type=pd.Timestamp, default=None)

    return app


# groups: symbol, frequency, field, start_date, end_date
FILE_REGEX = r"^([A-Z0-9\.]+)_Candlestick_([0-9]+_[MDHS])_(BID|ASK)_([0-9]{2}\.[0-9]{2}\.[0-9]{4})-([0-9]{2}\.[0-9]{2}\.[0-9]{4}).csv$"


def main():

    args = cli_app().parse_args()

    a = Arctic(args.arctic_uri)

    result = read_backfill(
        paths=args.path,
        arctic=a,
        dry_run=args.dry_run,
        universe=args.universe,
        provider=args.provider,
        clean=args.clean,
    )

    service = sampling.get_ig_service()

    sampling.download_instruments(
        index_col=None,
        epics=ASSET_MAPPING.values(),
        universe=args.universe,
        arctic=a,
    )

    if args.dry_run:
        print(result)
    print("finished")


@symbol_publisher(
    template="prices/{0}.{1}",
    symbol_prefix="{provider}.{universe}.",
    library_options=LibraryOptions(dynamic_schema=True),
)
def read_backfill(
    paths: List[Path],
    end_date: Optional[pd.Timestamp] = None,
    **kwargs,
):

    data_files = defaultdict(list)

    for path in paths:

        for file in os.listdir(path):

            if match := re.match(FILE_REGEX, file):

                symbol, frequency, field, start_date, end_date = match.groups()

                # need to group by symbol

                data_files[(field.lower(), symbol)].append(path / file)

    # symbol, ohlc, symbol
    #
    #
    def read_file(f):
        logger.warning("Reading %s", f)
        out = pd.read_csv(
            f,
            index_col=0,
            date_format="%d.%m.%Y %H:%M:%S.%f",
        )
        out.index = pd.to_datetime(out.index)
        if out.index.name == "Local time":
            out.index = out.index.tz_convert("utc")
        elif out.index.name == "Gmt time":
            out.index = out.index.tz_localize("GMT").tz_convert("utc")
        else:
            raise ValueError(out.index.name)
        return out.rename_axis("DateTime")

    result = pd.concat(
        (
            pd.concat(read_file(f) for f in files).query("~index.duplicated()")
            for files in data_files.values()
        ),
        axis=1,
        keys=data_files.keys(),
    ).reorder_levels([0, 2, 1], axis=1)

    result = result.mul(pd.Series(MULTIPLIERS), level=2, axis=1)

    result.rename(columns=ASSET_MAPPING, inplace=True)

    if end_date:

        result = result[result.index <= end_date]

    return (
        (result["bid"]["Open"], ("bid", "open")),
        (result["bid"]["High"], ("bid", "high")),
        (result["bid"]["Low"], ("bid", "low")),
        (result["bid"]["Close"], ("bid", "close")),
        (result["ask"]["Open"], ("ask", "open")),
        (result["ask"]["High"], ("ask", "high")),
        (result["ask"]["Low"], ("ask", "low")),
        (result["ask"]["Close"], ("ask", "close")),
        (((result["ask"]["Open"] + result["bid"]["Open"]) / 2), ("mid", "open")),
        (((result["ask"]["High"] + result["bid"]["High"]) / 2), ("mid", "high")),
        (((result["ask"]["Low"] + result["bid"]["Low"]) / 2), ("mid", "low")),
        (((result["ask"]["Close"] + result["bid"]["Close"]) / 2), ("mid", "close")),
    )


if __name__ == "__main__":

    import sys

    logging.getLogger(__name__).setLevel(logging.INFO)

    sys.argv.extend(
        [
            "--arctic-uri",
            "lmdb:///home/rory/dev/tradingo-plat/data/prod/tradingo.db",
            "--path",
            str(Path.home() / "dev" / "market-data" / "GAS"),
            str(Path.home() / "dev" / "market-data" / "USA500"),
            str(Path.home() / "dev" / "market-data" / "BRENT"),
            str(Path.home() / "dev" / "market-data" / "USTBOND"),
            str(Path.home() / "dev" / "market-data" / "COCOA"),
            # "--dry-run",
            "--clean",
            "--provider",
            "ig-trading",
            "--universe",
            "im-multi-asset",
            "--end-date",
            "2018-10-21 00:00:00+00:00",
        ]
    )

    main()



cimport cython
from libc.math cimport isnan, signbit
from cython.view cimport array as cvarray

cimport numpy as np
import numpy as np


cdef sign(float x):
    if signbit(x):
        return -1
    return 1


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef compute_backtest(
    float[:] trades,
    float[:] bid,
    float[:] ask,
    float[:] stop_limit,
    float[:] stop_loss,
    float[:] dividends,
):

    cdef size_t num_days = trades.shape[0]
    cdef size_t idx, idx_prev
 
    cdef float[:] unrealised_pnl = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] realised_pnl = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] total_pnl = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] net_investment = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] net_position = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] net_exposure = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] avg_open_price = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")
    cdef float[:] stop_trade = cvarray(shape=(num_days,), itemsize=sizeof(float), format="f")

    unrealised_pnl[0] = 0
    realised_pnl[0] = 0
    total_pnl[0] = 0
    net_investment[0] = 0
    net_position[0] = trades[0]
    if trades[0] != 0.0:
        avg_open_price[0] = bid[0] if trades[0] < 0 else ask[0]
    else:
        avg_open_price[0] = 0.0
    net_exposure[0] = trades[0] * avg_open_price[0]
    stop_trade[0] = 0

    # transient output variables
    cdef float m_net_position = 0
    cdef float m_avg_open_price = 0
    cdef float m_unrealised_pnl = 0
    cdef float m_total_pnl = 0
    cdef float m_realised_pnl = 0
    cdef float m_net_investment = 0
    cdef float m_loss_trade = 0
    cdef float m_net_exposure = 0

    # loop variables
    cdef float price
    cdef float trade_quantity
    cdef float trade_price
    cdef float c_stop_limit
    cdef float c_stop_loss

    for idx in range(1, num_days):

        idx_prev = idx - 1

        m_net_position = net_position[idx_prev]
        trade_quantity = trades[idx]
        c_stop_limit = stop_limit[idx]
        c_stop_loss = stop_loss[idx]
        m_loss_trade = 0
        price = (bid[idx] + ask[idx])/2

        if (
            (not isnan(c_stop_loss)) and (
                (m_net_position < 0.0 and price > c_stop_loss)
                or (m_net_position > 0.0 and price < c_stop_loss)
            )
        ) or (
            (not isnan(c_stop_limit)) and (
                (m_net_position < 0.0 and price < c_stop_limit)
                or (m_net_position > 0.0 and price > c_stop_limit)
            )
        ):
            m_loss_trade = -1 * m_net_position
            trade_quantity += m_loss_trade

        dividend = dividends[idx]
        m_unrealised_pnl = (price - avg_open_price[idx_prev]) * m_net_position 

        m_realised_pnl = realised_pnl[idx_prev]
        m_avg_open_price = avg_open_price[idx_prev] if m_net_position != 0 else 0.0

        if trade_quantity != 0 and not isnan(trade_quantity):

            trade_price = ask[idx] if trade_price > 0 else bid[idx]

            # net investment
            m_net_investment = max(
                m_net_investment, abs(m_net_position * m_avg_open_price)
            )
            # realized pnl
            if abs(m_net_position + trade_quantity) < abs(m_net_position):
                m_realised_pnl += (
                    (trade_price - m_avg_open_price) * abs(trade_quantity) * sign(m_net_position)
                )

            # avg open price
            if abs(m_net_position + trade_quantity) > abs(m_net_position):
                m_avg_open_price = (
                    (m_avg_open_price * m_net_position) + (trade_price * trade_quantity)
                ) / (m_net_position + trade_quantity)
            else:
                # Check if it is close-and-open
                if trade_quantity > abs(m_net_position):
                    m_avg_open_price = trade_price

        if dividend != 0 and not isnan(dividend):
            m_realised_pnl += m_net_position * dividend

        # total pnl
        m_total_pnl = m_realised_pnl + m_unrealised_pnl
        m_net_position += trade_quantity
        m_net_exposure = m_net_position * price

        unrealised_pnl[idx] = m_unrealised_pnl
        realised_pnl[idx] = m_realised_pnl
        total_pnl[idx] = m_total_pnl
        net_investment[idx] = m_net_investment
        net_position[idx] = m_net_position
        avg_open_price[idx] = m_avg_open_price
        stop_trade[idx] = m_loss_trade
        net_exposure[idx] = m_net_exposure

    return np.column_stack(
        (
            np.asarray(unrealised_pnl),
            np.asarray(realised_pnl),
            np.asarray(total_pnl),
            np.asarray(net_investment),
            np.asarray(net_position),
            np.asarray(net_exposure),
            np.asarray(avg_open_price),
            np.asarray(stop_trade),
        )
    )



import pandas as pd
import dateutil.tz

from tradingo.symbols import symbol_publisher
from tradingo.sampling import get_ig_service


@symbol_publisher(
    template="live/activity.{0}",
    symbol_prefix="{provider}.{universe}.",
)
def get_activity_history(
    from_date,
    to_date,
    svc=None,
    **kwargs,
):
    """
    get activtiy history and return dataframe per asset
    """

    svc = svc or get_ig_service()

    act = svc.fetch_account_activity_by_date(
        from_date=pd.Timestamp(from_date), to_date=pd.Timestamp(to_date)
    )
    act["DateTime"] = (
        pd.to_datetime(act["date"] + " " + act["time"])
        .tz_localize(dateutil.tz.tzlocal())
        .tz_convert("utc")
    )

    return tuple(
        (
            data.set_index("DateTime")
            .sort_index()
            .drop(["date", "time", "period", "epic", "marketName"], axis=1)
            .replace("-", np.nan)
            .astype(
                {
                    "size": float,
                    "stop": float,
                    "limit": float,
                },
                errors="ignore",
            ),
            (epic,),
        )
        for epic, data in iter(act.groupby("epic"))
    )



from pandas.plotting._matplotlib.style import get_standard_colors

from typing import List, Union

import matplotlib.axes
import pandas as pd


def plot_multi(
    data: pd.DataFrame,
    x: Union[str, None] = None,
    y: Union[List[str], None] = None,
    spacing: float = 0.1,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot multiple Y axes on the same chart with same x axis.

    Args:
        data: dataframe which contains x and y columns
        x: column to use as x axis. If None, use index.
        y: list of columns to use as Y axes. If None, all columns are used
            except x column.
        spacing: spacing between the plots
        **kwargs: keyword arguments to pass to data.plot()

    Returns:
        a matplotlib.axes.Axes object returned from data.plot()

    Example:
    >>> plot_multi(df, figsize=(22, 10))
    >>> plot_multi(df, x='time', figsize=(22, 10))
    >>> plot_multi(df, y='price qty value'.split(), figsize=(22, 10))
    >>> plot_multi(df, x='time', y='price qty value'.split(), figsize=(22, 10))
    >>> plot_multi(df[['time price qty'.split()]], x='time', figsize=(22, 10))

    See Also:
        This code is mentioned in https://stackoverflow.com/q/11640243/2593810
    """
    from pandas.plotting._matplotlib.style import get_standard_colors

    # Get default color style from pandas - can be changed to any other color list
    if y is None:
        y = data.columns

    # remove x_col from y_cols
    if x:
        y = [col for col in y if col != x]

    if len(y) == 0:
        return
    colors = get_standard_colors(num_colors=len(y))

    if "legend" not in kwargs:
        kwargs["legend"] = False  # prevent multiple legends

    # First axis
    ax = data.plot(x=x, y=y[0], color=colors[0], **kwargs)
    ax.set_ylabel(ylabel=y[0])
    lines, labels = ax.get_legend_handles_labels()

    for i in range(1, len(y)):
        # Multiple y-axes
        ax_new = ax.twinx()
        ax_new.spines["right"].set_position(("axes", 1 + spacing * (i - 1)))
        data.plot(ax=ax_new, x=x, y=y[i], color=colors[i % len(colors)], **kwargs)
        ax_new.set_ylabel(ylabel=y[i])

        # Proper legend position
        line, label = ax_new.get_legend_handles_labels()
        lines += line
        labels += label

    ax.legend(lines, labels, loc=0)
    return ax



import logging
import re
from typing import Optional

import pandas as pd
import numpy as np
import riskfolio as rf

import arcticdb as adb

from tradingo.symbols import lib_provider, symbol_provider, symbol_publisher


logger = logging.getLogger(__name__)


@lib_provider(model_signals="signals")
@symbol_provider(
    close="prices/close",
    vol="signals/vol_128",
    symbol_prefix="{provider}.{universe}.",
)
@symbol_provider(instruments="instruments/{universe}", no_date=True)
@symbol_publisher(
    template="{0}/{1}.{2}",
    symbol_prefix="{provider}.{universe}.{name}.",
)
def portfolio_construction(
    name: str,
    model_signals: adb.library.Library,
    close: pd.DataFrame,
    vol: pd.DataFrame,
    instruments: pd.DataFrame,
    provider: str,
    universe: str,
    signal_weights: dict,
    multiplier: float,
    default_weight: float,
    constraints: dict,
    vol_scale: bool,
    aum: float,
    instrument_weights: Optional[dict] = None,
    **kwargs,
):

    instrument_weights = instrument_weights or {}

    weights = pd.Series(multiplier, index=instruments.index)

    instruments["Symbol"] = instruments.index

    for key, weights_config in instrument_weights.items():

        if key not in instruments.columns:
            continue

        weights = weights * instruments.apply(
            lambda i: weights_config.get(i[key], default_weight),
            axis=1,
        )
        logger.info("Weights: %s", weights)

    if signal_weights:

        signal_value = (
            pd.concat(
                (
                    model_signals.read(
                        f"{provider}.{universe}.{signal}",
                        date_range=(close.index[0], close.index[-1]),
                    ).data
                    * weights
                    * signal_weight
                    for signal, signal_weight in signal_weights.items()
                ),
                keys=signal_weights,
                axis=1,
            )
            .transpose()
            .groupby(level=[1])
            .sum()
            .transpose()
            .ffill()
            .fillna(0)
        )

    else:
        signal_value = weights * pd.DataFrame(
            1,
            index=close.index,
            columns=close.columns,
        )

    if vol_scale:

        weights = weights.div(10 * vol)

    logger.info("signal_data: %s", signal_value)

    positions = signal_value.reindex_like(close).ffill()

    if constraints["long_only"]:
        positions[(positions < 0.0)] = 0.0

    pct_position = positions.div(positions.transpose().sum(), axis=0).fillna(0.0)
    for col in pct_position.columns:
        pct_position.loc[
            pct_position.index == pct_position.first_valid_index(), col
        ] = 0.0

    share_position = (pct_position * aum) / close

    return (
        (pct_position, ("portfolio", "raw", "percent")),
        (share_position, ("portfolio", "raw", "shares")),
        (positions, ("portfolio", "raw", "position")),
        (pct_position.round(decimals=2), ("portfolio", "rounded", "percent")),
        (share_position.round(), ("portfolio", "rounded", "shares")),
        (positions.round(), ("portfolio", "rounded", "position")),
        (signal_value, ("signals", "raw", "signal")),
    )


@symbol_provider(
    close="prices/{provider}.{universe}.close",
    factor_returns="prices/{factor_provider}.{factor_universe}.close",
)
@symbol_provider(instruments="instruments/{universe}", no_date=True)
@symbol_publisher(
    "portfolio/raw.percent",
    "portfolio/raw.shares",
    symbol_prefix="{provider}.{universe}.{name}.",
)
def portfolio_optimization(
    close: pd.DataFrame,
    factor_returns: pd.DataFrame,
    optimizer_config: dict,
    rebalance_rule: str,
    min_periods: int,
    aum: float,
    **kwargs,
):
    def get_weights(
        returns,
        factors,
    ):
        port = rf.Portfolio(returns=returns)

        port.assets_stats(method_mu="hist", method_cov="ledoit")
        port.lowerret = 0.00056488 * 1.5

        port.factors = factors

        port.factors_stats(
            method_mu="hist",
            method_cov="ledoit",
            feature_selection="PCR",
        )

        w = port.optimization(
            model="FM",
            rm="MV",
            obj="Sharpe",
            hist=False,
        )
        return (
            w.squeeze() if w is not None else pd.Series(np.nan, index=returns.columns)
        )

    asset_returns = close.pct_change().dropna()
    factor_returns = close.pct_change().dropna().reindex(asset_returns.index)

    data = []
    for i, _ in enumerate(asset_returns.index):

        if i < min_periods:
            data.append(
                pd.Series(np.nan, index=asset_returns.columns).to_frame().transpose()
            )
            continue

        ret_subset = asset_returns.iloc[:i]
        data.append(
            get_weights(ret_subset, factor_returns.loc[ret_subset.index])
            .to_frame()
            .transpose()
        )

    pct_position = pd.concat(data, keys=asset_returns.index).droplevel(1)
    share_position = (pct_position * aum) / close

    return (pct_position, share_position)


@symbol_provider(close="prices/adj_close", symbol_prefix="{provider}.{universe}.")
@symbol_publisher("prices/ivol", symbol_prefix="{provider}.{universe}.")
def instrument_ivol(close, provider, **kwargs):
    pct_returns = np.log(close / close.shift())

    ivols = []

    for symbol in pct_returns.columns:
        universe = pct_returns.drop(symbol, axis=1)

        def vol(uni):
            return (1 - (1 + uni).prod(axis=1).pow(1 / 100)).ewm(10).std()

        ivol = vol(pd.concat((universe, pct_returns[symbol]), axis=1)) - vol(universe)
        ivols.append(ivol.rename(symbol))

    return (pd.concat(ivols, axis=1).rename_axis("Symbol"),)


@symbol_provider(
    close="prices/close",
    symbol_prefix="{provider}.{universe}.",
)
@symbol_provider(instruments="instruments/{universe}", no_date=True)
@symbol_publisher(
    "portfolio/raw.percent",
    "portfolio/raw.shares",
    symbol_prefix="{provider}.{universe}.{name}.",
)
def position_from_trades(
    close: pd.DataFrame,
    instruments: pd.DataFrame,
    aum: float,
    trade_file: str,
    **kwargs,
):

    trades = (
        pd.read_csv(trade_file, parse_dates=["Date"])
        .dropna(axis=0, how="all")
        .sort_values(["Date"])
    )
    trades = trades[
        trades["Order type"].isin(["AtBest", "Quote and Deal"])
        & trades["Order status"].eq("Completed")
    ]
    trades["Ticker"] = (
        trades["Investment"].apply(
            lambda t: re.match(".*\(([A-Z]{4})\)", t).groups()[0]
        )
        + ".L"
    )
    position_shares = (
        trades.set_index(["Date", "Ticker"])
        .groupby(["Date", "Ticker"])
        .sum()
        .unstack()["My units"]
        .fillna(0.0)
        .cumsum()
        .reindex_like(
            close,
        )
        .ffill()
        .fillna(0.0)
    )

    position_pct = (position_shares * close) / aum
    return (
        position_pct,
        position_shares,
    )


# FIXME: Not idempotent
@symbol_provider(
    positions="portfolio/{stage}?as_of=-1",
    previous_positions="portfolio/{stage}?as_of=-2",
    symbol_prefix="{provider}.{universe}.{name}.",
)
@symbol_publisher(
    "trades/{stage}",
    symbol_prefix="{provider}.{universe}.{name}.",
)
def calculate_trades(
    name: str,
    stage: str,
    positions: pd.DataFrame,
    previous_positions: pd.DataFrame,
    **kwargs,
):

    logger.info(
        "Calculating %s trades for %s",
        stage,
        name,
    )

    return (
        (
            positions
            - previous_positions.reindex_like(positions, method="ffill").fillna(0.0)
        ).iloc[
            -1:,
        ],
    )



import string
import numpy as np
import pandas as pd
import pytest

from tradingo.api import Tradingo
from tradingo.utils import null_instruments


@pytest.fixture(scope="session")
def prices() -> pd.DataFrame:

    annual_std = 0.12

    returns = pd.DataFrame(
        np.random.normal(0, annual_std / np.sqrt(260), (100, 60)),
        index=pd.bdate_range(start="2024-01-03 00:00:00+00:00", periods=100),
        columns=[
            "".join(np.random.choice(list(string.ascii_uppercase), 4))
            for _ in range(60)
        ],
    )
    return (1 + returns).cumprod()


@pytest.fixture(scope="session")
def position(prices: pd.DataFrame):
    return pd.DataFrame(
        np.random.choice((1, 0, -1), prices.shape),
        index=prices.index,
        columns=prices.columns,
    ).cumsum()


@pytest.fixture(scope="session")
def tradingo(prices: pd.DataFrame, position: pd.DataFrame) -> Tradingo:
    t = Tradingo(
        name="test", provider="yfinance", universe="etfs", uri="mem://tradingo"
    )
    libraries = ["prices", "signals", "backtest", "portfolio", "instruments"]
    for library in libraries:
        t.create_library(library)
    t.instruments.etfs.update(null_instruments(prices.columns), upsert=True)
    t.prices.close.update(prices, upsert=True)
    t.prices.dividend.update(
        pd.DataFrame(0, index=prices.index, columns=prices.columns),
        upsert=True,
    )
    t.portfolio.model.raw.shares.update(position, upsert=True)
    return t



import argparse
import logging
from typing import Optional

import numpy as np
import pandas as pd

from trading_ig.rest import IGService
from tradingo.api import Tradingo
from tradingo.sampling import get_ig_service
from tradingo.symbols import symbol_provider, symbol_publisher


logger = logging.getLogger(__name__)


def close_position(deal_id, position, svc, size=None):
    direction = "BUY" if position.direction == "SELL" else "SELL"

    result = svc.close_open_position(
        deal_id=deal_id,
        direction=direction,
        epic=None,
        expiry=None,
        level=None,
        order_type="MARKET",
        size=size or abs(position["size"]),
        quote_id=None,
    )

    logger.info(result)


def close_all_open_position(positions, svc):

    epic_positions = positions

    for deal_id, position in epic_positions.iterrows():

        close_position(deal_id=deal_id, position=position, svc=svc)


def get_current_positions(
    service: IGService,
):

    all_positions = service.fetch_open_positions().set_index(["epic", "dealId"])
    all_positions["size"] = (
        all_positions["direction"].replace({"BUY": 1, "SELL": -1})
        * all_positions["size"]
    )
    return all_positions


def reduce_open_positions(
    service: IGService,
    epic: str,
    quantity: int,
):

    positions = get_current_positions(service).loc[epic]

    quantity_cxd = 0.0

    for deal_id, position in positions.sort_values("size").iterrows():

        to_cancel = min(position["size"], quantity - quantity_cxd)

        close_position(
            deal_id=deal_id,
            position=position,
            svc=service,
            size=to_cancel,
        )

        quantity_cxd += to_cancel

        if quantity_cxd >= quantity:
            break


def cli_app():
    app = argparse.ArgumentParser()

    app.add_argument("--arctic-uri", required=True)
    app.add_argument("--name", required=True)
    app.add_argument("--universe", required=True)
    app.add_argument("--provider", required=True)
    app.add_argument("--portfolio-name", required=True)
    app.add_argument("--stage", default="rounded.position")

    return app


def get_currency(instrument: pd.Series):
    if "$" in instrument.name:
        return "USD"
    elif "£" in instrument.name:
        return "GBP"
    elif "€" in instrument.name:
        return "EUR"
    return "GBP"


# TODO: #21 - symbol_publisher
@symbol_provider(
    instruments="instruments/{universe}",
    no_date=True,
)
@symbol_provider(
    target_positions="portfolio/{name}.{stage}",
    symbol_prefix="{provider}.{universe}.",
)
def adjust_position_sizes(
    stage: str,
    instruments: pd.DataFrame,
    target_positions: pd.DataFrame,
    universe: str,
    provider: str,
    name: str,
    service: Optional[IGService] = None,
    **kwargs,
):

    logger.info(
        "Adjusting name=%s universe=%s provider=%s stage=%s",
        name,
        universe,
        provider,
        stage,
    )

    service = service or get_ig_service()
    current_positions = get_current_positions(service)

    current_sizes = (
        current_positions.groupby("epic")["size"]
        .sum()
        .reindex(target_positions.columns)
        .fillna(0.0)
    )

    for epic in target_positions:
        latest_target = target_positions[epic].iloc[-1]
        current_position = current_sizes.loc[epic]

        # changing sides, close existing
        if current_position and np.sign(current_position) != np.sign(latest_target):
            logger.info("Closing open position of %s for %s", current_position, epic)
            close_all_open_position(current_positions.loc[epic], service)
            current_position = 0.0

        # increasing position
        if abs(current_position) < abs(latest_target):

            target = abs(latest_target) - abs(current_position)
            side = "BUY" if latest_target > 0 else "SELL"

            logger.info(
                "Increasing target position from %s to %s - %s for %s",
                current_position,
                latest_target,
                side,
                epic,
            )

            result = service.create_open_position(
                direction=side,
                currency_code=get_currency(instruments.loc[epic]),
                order_type="MARKET",
                expiry=instruments.loc[epic].expiry,
                size=target,
                epic=epic,
                force_open=False,
                guaranteed_stop=False,
                level=None,
                limit_distance=None,
                limit_level=None,
                quote_id=None,
                stop_distance=None,
                trailing_stop=None,
                trailing_stop_increment=None,
                stop_level=None,
            )

            logger.info(result)

        elif abs(current_position) > abs(latest_target):

            reduce_by = abs(current_position - latest_target)

            logger.info(
                "Reducing open positions by %s from %s to %s for %s",
                reduce_by,
                current_position,
                latest_target,
                epic,
            )

            reduce_open_positions(
                service,
                epic=epic,
                quantity=reduce_by,
            )
        else:
            logger.info(
                "%s matches target, nothing to do for %s.",
                current_position,
                epic,
            )


def main():

    args = cli_app().parse_args()

    service = get_ig_service()



from numpy import divide
import pytest
import pandas as pd

from tradingo import backtest

from tradingo.test.utils import close_position


def test_backtest_integration(benchmark, tradingo):

    bt = benchmark(
        backtest.backtest,
        bid="prices/close",
        ask="prices/close",
        dividends="prices/dividend",
        start_date=pd.Timestamp("2018-01-01 00:00:00+00:00"),
        end_date=pd.Timestamp.now("utc"),
        name="model",
        stage="raw.shares",
        config_name="test",
        provider="yfinance",
        universe="etfs",
        dry_run=True,
        arctic=tradingo,
    )


@pytest.mark.parametrize(
    "prices_,portfolio_,unrealised_pnl,realised_pnl",
    [
        (
            pd.DataFrame(
                1 + 0.01,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            pd.DataFrame(
                1,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            "(prices_*portfolio_).squeeze().diff()",
            "pd.Series(0.0, index=prices_.index)",
        ),
        (
            pd.DataFrame(
                1 + 0.01,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            pd.DataFrame(
                range(0, 260),
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ),
            "(prices_.diff()*portfolio_.shift()).squeeze()",
            "pd.Series(0.0, index=prices_.index)",
        ),
        (
            pd.DataFrame(
                1 + 0.01,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            close_position(
                pd.DataFrame(
                    1,
                    index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                    columns=["ABCD"],
                ).cumprod(),
            ),
            "(prices_.diff()*portfolio_.shift()).squeeze()",
            "pd.Series([*(0 for _ in range(0, 259)), 12.280985478055433], index=prices_.index)",
        ),
    ],
)
def test_backtest_smoke(tradingo, prices_, portfolio_, unrealised_pnl, realised_pnl):

    dividends = pd.DataFrame(0, columns=prices_.columns, index=prices_.index)

    bt = backtest.backtest(
        portfolio=portfolio_,
        bid_close=prices_,
        ask_close=prices_,
        dividends=dividends,
        name="test",
        config_name="test",
        dry_run=True,
        start_date=portfolio_.index[0],
        end_date=portfolio_.index[-1],
        stage="raw",
        provider="yfinance",
        universe="etfs",
        arctic=tradingo,
    )

    actual_unrealised = bt["backtest/instrument.unrealised_pnl"].squeeze().diff()
    expected_unrealised = (
        eval(unrealised_pnl) if isinstance(unrealised_pnl, str) else unrealised_pnl
    )

    pd.testing.assert_series_equal(
        actual_unrealised,
        expected_unrealised.astype("float32"),
        check_names=False,
        check_freq=False,
        rtol=1e-4,
    )
    actual_realised = (
        bt["backtest/instrument.realised_pnl"].squeeze().diff().fillna(0.0)
    )
    expected_realised = (
        eval(realised_pnl) if isinstance(realised_pnl, str) else realised_pnl
    )
    pd.testing.assert_series_equal(
        actual_realised,
        expected_realised.astype("float32"),
        check_names=False,
        check_freq=False,
        rtol=1e-4,
    )


if __name__ == "__main__":
    pytest.main()



from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy as np

setup(
    ext_modules=cythonize(
        [
            Extension(
                name="tradingo._backtest",
                sources=["src/tradingo/_backtest.pyx"],
            ),
        ],
        gdb_debug=False,
    ),
    include_dirs=[
        np.get_include(),
    ],
)
